package de.unima.crumbtrail.pruner;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Lists;
import com.google.common.collect.Multimap;
import de.unima.crumbtrail.utility.ScreenUpdate;

import it.uniroma1.lcl.jlt.util.DoubleCounter;
import it.uniroma1.lcl.jlt.util.Files;
import it.uniroma1.lcl.jlt.util.IntegerCounter;
import java.io.BufferedWriter;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;

import java.util.HashSet;
import java.util.List;
import java.util.Map;

import java.util.Set;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadPoolExecutor;
import org.jgrapht.alg.BellmanFordShortestPath;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;
import org.jgrapht.alg.DijkstraShortestPath;

/**
 * NOTE: this is a prototype implementation 
 *       we are working hard to produce and release an optimized and documented version soon. 
 * @author sfaralli
 */
public class CrumbTrail {

    public static final int bins = 6;
    public static final int BF = 0;
    public static final int SPmethod = BF;
    public static final int upuntilsave = 100;
    public static final int monothreadlimit = 50;

    public static int dataoutcounter = 0;

    /**
     *
     * @param datatout
     * @param forcetestcycle
     * @param alsopostponed
     * @param optimized: when true, apply optimization on level outgoingedges
     * @param additional: true, remove unreacheable nodes vengono eliminati
     * @param noisyGraph: the input noisy graph and the output pruned graph
     * @param rootLabel: the label of the rootLabel node
     * @param leafNodes: the set of leaves
     * @param intermediate: the set of intermediate
     * @param filename: lof file
     * @param label: label used to distinguish the run
     * @param label2: label used to distinguish the run
     * @param method: label used to distinguish the run
     * @param removeOnlyCycles: if true remove only cycles
     * @param debug: if true logs additional debug inforamtion
     * @throws IOException
     * @throws InterruptedException
     */
    public static void prune(String datatout, boolean forcetestcycle, boolean alsopostponed, boolean optimized, boolean additional, DefaultDirectedGraph<String, DefaultEdge> noisyGraph, String rootLabel, Set<String> leafNodes, Set<String> intermediate, String filename, String label, String label2, String method, boolean removeOnlyCycles, boolean debug) throws IOException, InterruptedException {

        IntegerCounter<String> postponedcounts = new IntegerCounter<>();
        Set<DefaultEdge> testededges = new HashSet<>();
        HashMultimap<String, String> testedpostponed = HashMultimap.create();
        HashMultimap<Integer, String> level2nodes = HashMultimap.create();
        HashMultimap<String, String> node2pusher = HashMultimap.create();
        //if additional is true remove unreacheable nodes 
        if (!additional) {
            Set<String> alltheothers = new HashSet<>(noisyGraph.vertexSet());
            alltheothers.remove(rootLabel);
            BellmanFordShortestPath bfcheck = new BellmanFordShortestPath(noisyGraph, rootLabel);

            for (String s : alltheothers) {

                Double cost = bfcheck.getCost(s);
                if (cost.equals(Double.POSITIVE_INFINITY)) {
                    System.out.println("********* UNREACHABLE **************" + s);
                    noisyGraph.removeVertex(s);
                    leafNodes.remove(s);
                    intermediate.remove(s);
                }
            }
        }
        // logs the basic information about the noisy graph
        BufferedWriter bw = Files.getBufferedWriter(filename);
        bw.write("$|G.V|: " + noisyGraph.vertexSet().size() + "\n");
        bw.write("$|G.E|: " + noisyGraph.edgeSet().size() + "\n");
        if (debug) {
            for (DefaultEdge de : noisyGraph.edgeSet()) {
                bw.write("$ACTUAL: " + de + "\n");
            }
        }

        // remove unreacheable nodes form the Intermediate and the leaves sets
        removeUnConnected(noisyGraph, rootLabel, leafNodes, intermediate);

        // create a set with the root node
        Set<String> roots = new HashSet<>();
        roots.add(rootLabel);

        // remove inedges of rootLabel
        Set<DefaultEdge> toedgre = new HashSet<>();
        for (DefaultEdge de : noisyGraph.incomingEdgesOf(rootLabel)) {
            toedgre.add(de);
        }

        Set<String> analyzed = new HashSet<>();
        Set<String> removedPending = new HashSet<>();
        Set<String> removedNoisy = new HashSet<>();
        Set<DefaultEdge> removedEdges = new HashSet<>();
        Multimap<Integer, String> pushed = HashMultimap.create();

        saveState(datatout, "Initial noisy", toGraph(new HashSet<String>(), noisyGraph, pushed, analyzed, removedPending, removedNoisy, removedEdges, roots, leafNodes, intermediate));

        for (DefaultEdge de : toedgre) {
            removedEdges.add(de);
            saveState(datatout, "Preliminar incoming edge of root", toGraph(new HashSet<String>(), noisyGraph, pushed, analyzed, removedPending, removedNoisy, removedEdges, roots, leafNodes, intermediate));
            bw.write("$REMOVE-EDGE INCOMING OF ROOT: " + de + "\n");
            noisyGraph.removeAllEdges(noisyGraph.getEdgeSource(de), noisyGraph.getEdgeTarget(de));
            System.out.println("Removing edge:" + de);
        }
        // removes self-loops
        System.out.println("Remove self loops...");
        for (String s : noisyGraph.vertexSet()) {
            DefaultEdge de = noisyGraph.getEdge(s, s);
            if (de != null) {
                removedEdges.add(de);
                saveState(datatout, "Preliminar self loop", toGraph(new HashSet<String>(), noisyGraph, pushed, analyzed, removedPending, removedNoisy, removedEdges, roots, leafNodes, intermediate));
                noisyGraph.removeAllEdges(noisyGraph.getEdgeSource(de), noisyGraph.getEdgeTarget(de));
                System.out.println("Removed self loop on:" + s);
                bw.write("$REMOVE EDGE SELF-LOOP: " + de + "\n");

            }
        }
        // remove isolated nodes
        removeIsolated(noisyGraph);

        // additional sanity checks on the set of leaves 
        System.out.println("Clean the goods....");
        for (String g : new HashSet<>(leafNodes)) {
            if (!noisyGraph.containsVertex(g)) {
                System.out.println(g + " is not good!");
                bw.write(g + " is not good!\n");
                leafNodes.remove(g);
            } else if (noisyGraph.incomingEdgesOf(g).isEmpty() && noisyGraph.outgoingEdgesOf(g).isEmpty()) {
                System.out.println(g + " is not good ISOLATED!");
                bw.write(g + " is not good ISOLATED!\n");
                noisyGraph.removeVertex(g);
                leafNodes.remove(g);
                analyzed.remove(g);
            }
        }

        // log
        int max = 0;

        // struttura temporanea 
        Set<String> saved = new HashSet<>();

        System.out.println("removecycles out of");
        //prima fase del pruning: si rimuovono i cicli passanti per gli outgoing edges dei protetti
        /*BellmanFordShortestPath bfoot = new BellmanFordShortestPath(noisyGraph, rootLabel);*/
        Set<String> KK = new HashSet<>();
        KK.add(rootLabel);
        for (String kc : leafNodes) {
            for (DefaultEdge de : new HashSet<>(noisyGraph.outgoingEdgesOf(kc))) {
                DefaultEdge removed = removeCyclesOverEdge(additional, 0, "OUTOGING-EDGE OF P", noisyGraph.getEdgeTarget(de), kc, noisyGraph, rootLabel, /*bfoot,*/ bw, KK, new HashSet<String>(), debug, leafNodes);
                if (removed != null) {
                    removedEdges.add(removed);
                    saveState(datatout, "Preliminar outgoing edge of p in P", toGraph(new HashSet<String>(), noisyGraph, pushed, analyzed, removedPending, removedNoisy, removedEdges, roots, leafNodes, intermediate));
                }
                testededges.add(de);
            }
        }

        System.out.println("REMOVING PENDING....");
        boolean removed0;
        do {
            removed0 = false;
            Set<String> toremove = new HashSet<>();
            for (String h : noisyGraph.vertexSet()) {
                if (!leafNodes.contains(h) && noisyGraph.outgoingEdgesOf(h).size() == 0) {
                    removed0 = true;
                    toremove.add(h);
                    removedPending.add(h);
                    removedEdges.addAll(noisyGraph.incomingEdgesOf(h));
                    removedEdges.addAll(noisyGraph.outgoingEdgesOf(h));
                }
            }
            for (String d : toremove) {

                noisyGraph.removeVertex(d);
                intermediate.remove(d);
                analyzed.remove(d);
                if (debug) {
                    bw.write("$REMOVE-NODE PENDING LEAF: " + d + "\n");
                }
                System.out.println("Removing fake leaf:" + d);
            }
        } while (removed0);
        do {
            removed0 = false;
            Set<String> toremove = new HashSet<>();
            for (String h : noisyGraph.vertexSet()) {
                if (!h.equals(rootLabel) /*&& !leafNodes.contains(h) */ /*&& !leafNodes.contains(h) */ && noisyGraph.incomingEdgesOf(h).isEmpty()) {
                    removed0 = true;
                    toremove.add(h);
                    removedPending.add(h);
                    removedEdges.addAll(noisyGraph.incomingEdgesOf(h));
                    removedEdges.addAll(noisyGraph.outgoingEdgesOf(h));
                }
            }
            for (String d : toremove) {

                noisyGraph.removeVertex(d);
                intermediate.remove(d);
                leafNodes.remove(d);
                analyzed.remove(d);
                bw.write("$REMOVE-NODE PENDING ROOT: " + d + "\n");
                System.out.println("Removing fake root:" + d);
            }
        } while (removed0);

        saveState(datatout, "Remove pending", toGraph(new HashSet<String>(), noisyGraph, pushed, analyzed, removedPending, removedNoisy, removedEdges, roots, leafNodes, intermediate));

        leafNodes.removeAll(intermediate);

        if (debug) {
            bw.write("$GROUND: " + leafNodes + " " + leafNodes.size() + "\n");
            bw.write("$INTERMEDIATE: " + intermediate + " " + intermediate.size() + "\n");
            System.out.println("GOOD:" + leafNodes.size() + " " + leafNodes);
            System.out.println("K:" + intermediate.size() + " " + intermediate);

        } else {
            bw.write("$GROUND: " + leafNodes + " " + leafNodes.size() + "\n");
            bw.write("$INTERMEDIATE: " + intermediate + " " + intermediate.size() + "\n");
            System.out.println("GOOD:" + leafNodes.size());
            System.out.println("K:" + intermediate.size());

        }
        // il primo livello contiene i nodi leaves
        int leveli = 0;
        List<String> level = new ArrayList<String>();
        level.addAll(leafNodes);

        List<String> nextlevel = null;
        boolean somethingchanged;
        saved.addAll(level);
        analyzed.addAll(level);

        HashMultimap<String, String> passing = HashMultimap.create();
        HashMultimap<String, String> passingk = HashMultimap.create();
        IntegerCounter<String> node2goods = new IntegerCounter<String>();
        IntegerCounter<String> node2ks = new IntegerCounter<String>();

        Set<String> essential = new HashSet<>();
        for (String s : noisyGraph.vertexSet()) {
            postponedcounts.count(s, 0);
        }
        // main iteration cycle
        do {

            if (debug) {
                for (DefaultEdge de : noisyGraph.edgeSet()) {
                    bw.write("$ACTUAL: " + de + "\n");
                }
                bw.flush();
            }

            if (debug) {
                bw.write("$LEVEL:" + level + " " + level.size() + "\n");
                bw.flush();
            } else {
                bw.write("$LEVEL:" + level.size() + "\n");
                bw.flush();
            }

            bw.write("$NEW LEVEL" + "\n");
            System.out.println("NEW LEVEL");

            boolean firstlevel = true;
            do {

                somethingchanged = false;
                // get the parents of nodes
                if (firstlevel) {
                    for (int plevel = 0; plevel <= leveli; plevel++) {
                        if (pushed.containsKey(plevel)) {
                            pushed.removeAll(plevel);
                        }
                    }
                    int maxs = 0;
                    for (int ik : pushed.keySet()) {
                        if (ik > maxs) {
                            maxs++;
                        }
                    }
                    leveli = leveli + 1;

                    nextlevel = new ArrayList<>();
                    saveState(datatout, "Iteration:" + leveli, toGraph(new HashSet<String>(), noisyGraph, pushed, analyzed, removedPending, removedNoisy, removedEdges, roots, leafNodes, intermediate));

                    // updates the list of intermediate
                    for (String a : new HashSet<>(intermediate)) {
                        if (saved.contains(a)) {
                            if (debug) {
                                bw.write("$TOGROUND:" + a + "\n");
                                System.out.println(a + " passed K->G and added to level cause is leafleaf");
                            }
                            level.add(a);
                            leafNodes.add(a);
                            intermediate.remove(a);
                        }
                    }
                    if (debug) {
                        bw.write("$GROUND: " + leafNodes + " " + leafNodes.size() + "\n");
                        bw.write("$INTERMEDIATE: " + intermediate + " " + intermediate.size() + "\n");
                    }
                    saveState(datatout, "GroundVSIntermediate:" + leveli, toGraph(new HashSet<String>(nextlevel), noisyGraph, pushed, analyzed, removedPending, removedNoisy, removedEdges, roots, leafNodes, intermediate));
                    int z = 0;
                    ScreenUpdate.cleanScreen();
                    if (debug) {
                        bw.write("$BUILDINGLEVEL:" + leveli + "\n");
                    }
                    // cnstruction of the level 
                    ScreenUpdate.printText("Building: level " + leveli, 1, 1);
                    Set<String> Kr = new HashSet<>(intermediate);
                    Kr.add(rootLabel);

                    for (String s : level) {
                        z++;
                        ScreenUpdate.printText(z + "/" + level.size(), 3, 1);
                        if (noisyGraph.containsVertex(s)) {

                            for (DefaultEdge de : new HashSet<DefaultEdge>(noisyGraph.incomingEdgesOf(s))) {

                                String so = noisyGraph.getEdgeSource(de);

                                if (!analyzed.contains(so)) {
                                    Set<String> values = new HashSet<>(pushed.values());
                                    values.removeAll(pushed.get(leveli));
                                    if (!values.contains(so)) {
                                        BellmanFordShortestPath<String, DefaultEdge> bf = new BellmanFordShortestPath(noisyGraph, so);
                                        boolean found = false;
                                        for (String push : values) {
                                            Double costo = bf.getCost(push);
                                            if (costo != Double.POSITIVE_INFINITY) {
                                                found = true;
                                                break;
                                            }
                                        }
                                        if (!found) {
                                            node2pusher.put(so, s);
                                            nextlevel.add(so);
                                        }
                                    }
                                }
                            }
                        }
                    }//each s in level 
                    saveState(datatout, "Bulding Layer from lower layer:" + leveli, toGraph(new HashSet<String>(nextlevel), noisyGraph, pushed, analyzed, removedPending, removedNoisy, removedEdges, roots, leafNodes, intermediate));
                    z = 0;
                    ScreenUpdate.cleanScreen();
                    if (debug) {
                        bw.write("$LEVELAFTERBUILDINGLEVEL:" + nextlevel + "\n");
                        bw.write("$BUILDINGLEVEL-RESCHEDUED:" + leveli + "\n");
                    }
                    // cnstruction of the level from the previously postponed nodes
                    int max2 = 0;
                    for (Integer i : pushed.keySet()) {
                        if (i > max2) {
                            max2 = i;
                        }
                    }

                    do {
                        ScreenUpdate.printText("Building pushed: level " + leveli, 1, 1);
                        HashSet<String> levis = new HashSet<>(pushed.get(leveli));
                        int size = levis.size();
                        for (String s : levis) {

                            z++;
                            ScreenUpdate.printText(leveli + " " + z + "/" + size, 3, 1);
                            if (noisyGraph.containsVertex(s)) {
                                if (!analyzed.contains(s)) {

                                    nextlevel.add(s);
                                    node2pusher.put(s, "posponed level:" + leveli);
                                    for (int i : new HashSet<>(pushed.keySet())) {
                                        pushed.remove(i, s);
                                    }
                                }
                            } else {

                                for (int i : new HashSet<Integer>(pushed.keySet())) {
                                    pushed.remove(i, s);
                                }
                            }
                        }

                        if (nextlevel.isEmpty()) {
                            leveli++;
                        }
                        if (leveli > max2) {
                            break;
                        }
                    } while (nextlevel.isEmpty());

                    saveState(datatout, "Bulding Layer from postponed:" + leveli, toGraph(new HashSet<>(nextlevel), noisyGraph, pushed, analyzed, removedPending, removedNoisy, removedEdges, roots, leafNodes, intermediate));

                    Set<String> hs = new HashSet<>();
                    hs.addAll(nextlevel);
                    nextlevel.clear();
                    nextlevel.addAll(hs);
                    hs = null;
                    if (debug) {
                        bw.write("$LEVELAFTERBUILDINGLEVEL-RESCHEDULED:" + nextlevel + "\n");
                    }

                    System.out.println("REMOVING cycles on Layer's edges");
                    HashSet<String> border = new HashSet<>();
                    for (String s : nextlevel) {
                        for (DefaultEdge de : new HashSet<>(noisyGraph.incomingEdgesOf(s))) {
                            String source = noisyGraph.getEdgeSource(de);
                            if (!nextlevel.contains(source)) {
                                border.add(s);
                            }

                        }
                    }

                    HashSet<DefaultEdge> visitededges = new HashSet<DefaultEdge>();
                    while (!border.isEmpty()) {
                        HashSet<String> newborder = new HashSet<>();
                        for (String s : border) {
                            for (DefaultEdge de : new HashSet<>(noisyGraph.outgoingEdgesOf(s))) {
                                if (visitededges.contains(de)) {
                                    continue;
                                }
                                visitededges.add(de);
                                String target = noisyGraph.getEdgeTarget(de);
                                if (nextlevel.contains(target)) {
                                    newborder.add(target);
                                }
                                if (!testededges.contains(de)) {
                                    DefaultEdge removed = removeCyclesOverEdge(additional, 0, "OUTOGING-EDGE OF P", noisyGraph.getEdgeTarget(de), noisyGraph.getEdgeSource(de), noisyGraph, rootLabel, bw, KK, new HashSet<String>(), debug, leafNodes);
                                    if (removed != null) {
                                        removedEdges.add(removed);
                                        saveState(datatout, "Removing cycle inside the layer:" + leveli, toGraph(new HashSet<String>(nextlevel), noisyGraph, pushed, analyzed, removedPending, removedNoisy, removedEdges, roots, leafNodes, intermediate));
                                    }
                                    testededges.add(de);
                                }
                            }
                        }
                        border.clear();
                        border.addAll(newborder);
                    }

                    System.out.println("REMOVING OUTGOING");
                    for (String s : nextlevel) {
                        for (DefaultEdge de : new HashSet<>(noisyGraph.outgoingEdgesOf(s))) {
                            String target = noisyGraph.getEdgeTarget(de);
                            if (level.contains(target)) {
                                for (DefaultEdge de2 : new HashSet<>(noisyGraph.outgoingEdgesOf(s))) {
                                    if (!testededges.contains(de2)) {
                                        DefaultEdge removed = removeCyclesOverEdge(additional, 0, "OUTOGING-EDGE OF P", noisyGraph.getEdgeTarget(de2), noisyGraph.getEdgeSource(de2), noisyGraph, rootLabel, bw, KK, new HashSet<String>(), debug, leafNodes);
                                        if (removed != null) {
                                            removedEdges.add(removed);
                                            saveState(datatout, "Removing cycle inside the layer:" + leveli, toGraph(new HashSet<String>(nextlevel), noisyGraph, pushed, analyzed, removedPending, removedNoisy, removedEdges, roots, leafNodes, intermediate));
                                        }
                                        testededges.add(de);
                                    }
                                }
                                break;
                            }
                        }
                    }

                    // esamina i nodi del livello e determina i nodi da postporre a livelli successivi 
                    Set<String> pumped = new HashSet<>();
                    Map<String, Integer> n2l = new HashMap<String, Integer>();
                    int numerobins = Math.max(nextlevel.size() / bins, 1);

                    List<List<String>> partition = new ArrayList<List<String>>();
                    if (nextlevel.size() >= monothreadlimit) 
                    {
                        if (numerobins > 0) 
                        {
                            partition = Lists.partition(nextlevel, numerobins);
                        }
                        numerobins = partition.size();
                    } else 
                    {
                        partition.add(nextlevel);
                    }
                    numerobins = partition.size();
                    ThreadPoolExecutor spe = (ThreadPoolExecutor) Executors.newFixedThreadPool(numerobins);

                    if (SPmethod == BF) {
                        List<LevelThread> lth = new ArrayList<LevelThread>();
                        for (int nb = 0; nb < numerobins; nb++) {
                            List<String> subl = new ArrayList<String>(partition.get(nb));
                            LevelThread tr = new LevelThread(nb, leveli, subl, pushed, n2l, pumped, analyzed, nextlevel, noisyGraph, testedpostponed);
                            spe.submit(tr);
                            lth.add(tr);
                        }
                        bw.flush();
                        ScreenUpdate.cleanScreen();
                        ScreenUpdate.printText("PUSHING:level " + leveli, 1, 1);
                        while (true) {

                            boolean somethingstill = false;
                            //    if (leveli<=2)
                            for (int i = 0; i < numerobins; i++) {
                                LevelThread ltt = lth.get(i);
                                if (!ltt.done) {

                                    String text = i + " " + ltt.getActual() + "/" + ltt.getTotal() + " - " + ltt.getOverall();
                                    ScreenUpdate.printText(text, i + 3, 1);
                                    somethingstill = true;

                                } else {

                                    String text = i + " ******  done! *******";
                                    ScreenUpdate.printText(text, i + 3, 1);

                                }
                            }
                            if (!somethingstill) {
                                break;
                            }
                            try {
                                Thread.sleep(100);
                            } catch (Exception e) {

                            }
                        }
                        spe.shutdown();
                    }
                    if (debug) {
                        bw.write(leveli + " PUMPED:" + pumped.size() + "\n");
                        bw.flush();
                    }
                    // rimuove dal livello in costruzione i nodi che sono stati postponed

                    nextlevel.removeAll(pumped);
                    for (String s : pumped) {
                        postponedcounts.count(s);
                    }
                    level2nodes.putAll(leveli, nextlevel);
                    saveState(datatout, "After postponing:" + leveli, toGraph(new HashSet<String>(nextlevel), noisyGraph, pushed, analyzed, removedPending, removedNoisy, removedEdges, roots, leafNodes, intermediate));

                    
                    if (debug) {
                        bw.write("$RESCHEDULED: " + pushed + "\n");
                        System.out.println(leveli + " PUSHED:" + pushed + "\n");
                    }

                    bw.flush();

                    // elimina i nodi ridondanti
                    hs = new HashSet<>();
                    hs.addAll(nextlevel);
                    nextlevel.clear();
                    nextlevel.addAll(hs);
                    // Collections.sort(nextlevel);
                    hs = null;
                    passing.clear();
                    passingk.clear();
                    node2goods.clear();
                    node2ks.clear();
                    ScreenUpdate.cleanScreen();
                    ScreenUpdate.printText("Precomputing the passing list: level " + leveli, 1, 1);

                    if (debug) {
                        System.out.println("ANALYZED:" + analyzed);
                        System.out.println("SAVED:" + saved);
                        System.out.println("NEXTLEVEL:" + nextlevel);
                    }

                    // aggiornata la lista dei nodi intermedi
                    for (String a : analyzed) {
                        if (intermediate.contains(a)) {
                            System.out.println(a + " passed K->G");
                            leafNodes.add(a);
                            intermediate.remove(a);
                        }
                    }

                    // calcola l'insieme dei nodi essenziali
                    numerobins = Math.max(nextlevel.size() / bins, 1);

                    partition = new ArrayList<List<String>>();

                    if (nextlevel.size() >= monothreadlimit) {
                        if (numerobins > 0) {
                            partition = Lists.partition(nextlevel, numerobins);
                        }
                        numerobins = partition.size();
                    } else {
                        //partition=Lists.partition(nextlevel, nextlevel.size());
                        partition.add(nextlevel);

                    }
                    numerobins = partition.size();
                    spe = (ThreadPoolExecutor) Executors.newFixedThreadPool(numerobins);

                    List<PassingThread> lth = new ArrayList<PassingThread>();
                    for (int nb = 0; nb < numerobins; nb++) {
                        List<String> subl = new ArrayList<String>(partition.get(nb));
                        PassingThread tr = new PassingThread(nb, leveli, subl, passing, passingk, noisyGraph, leafNodes, intermediate, rootLabel, analyzed, essential, node2goods, node2ks);
                        spe.submit(tr);
                        lth.add(tr);
                    }

                    bw.flush();
                    ScreenUpdate.cleanScreen();
                    ScreenUpdate.printText("PASSING:level " + leveli, 1, 1);
                    while (true) {
                        boolean somethingstill = false;
                        for (int i = 0; i < numerobins; i++) {
                            PassingThread ltt = lth.get(i);
                            if (!ltt.done) {

                                String text = i + " " + ltt.getActual() + "/" + ltt.getTotal() + " - " + ltt.getOverall();
                                ScreenUpdate.printText(text, i + 3, 1);
                                somethingstill = true;

                            } else {

                                String text = i + " ******  done! *******";
                                ScreenUpdate.printText(text, i + 3, 1);

                            }

                        }
                        if (!somethingstill) {
                            break;
                        }
                        try {
                            Thread.sleep(200);
                        } catch (Exception e) {

                        }
                    }
                    spe.shutdown();

                    firstlevel = false;
                    // salva il livello
                    BufferedWriter bwl;
                    if (!nextlevel.isEmpty() && (leveli <= upuntilsave || leveli % 10 == 0)) {
                        bwl = Files.getBufferedWriter("data/piero/" + label2 + "_level_" + leveli + ".tsv");
                        for (String sbw : nextlevel) {
                            bwl.write(sbw + "\n");
                        }
                        bwl.close();
                    }
                    // salva i pushed
                    if (leveli <= upuntilsave || leveli % 10 == 0) {
                        bwl = Files.getBufferedWriter("data/piero/" + label2 + "_level__pushed_" + leveli + ".tsv");
                        if (pushed.size() > 0) {
                            for (Integer i : pushed.keySet()) {
                                if (pushed.get(i) != null && !pushed.get(i).isEmpty()) {
                                    for (String sbw : pushed.get(i)) {
                                        bwl.write(i + "\t" + sbw + "\n");
                                    }
                                }
                            }
                        }
                        bwl.close();
                    }
                    if (!removeOnlyCycles && passing.size() > 0 && (leveli <= upuntilsave || leveli % 10 == 0)) {
                        bwl = Files.getBufferedWriter("data/piero/" + label2 + "_level__passing_" + leveli + ".tsv");
                        for (String s : passing.keySet()) {
                            bwl.write(s + "\t");
                            for (String sbw : passing.get(s)) {
                                bwl.write(sbw + "\t");
                            }
                            bwl.write("\n");
                        }
                        bwl.close();
                    }
                    // come in precedenza rimuove i nodi pending
                    System.out.println("REMOVING PENDING....");

                    boolean removed;
                    do {
                        removed = false;
                        Set<String> toremove = new HashSet<>();
                        for (String h : noisyGraph.vertexSet()) {
                            if (!leafNodes.contains(h) && noisyGraph.outgoingEdgesOf(h).isEmpty()) {
                                removed = true;
                                toremove.add(h);
                            }
                        }

                        for (String d : toremove) {
                            removedEdges.addAll(noisyGraph.incomingEdgesOf(d));
                            removedEdges.addAll(noisyGraph.outgoingEdgesOf(d));
                            noisyGraph.removeVertex(d);
                            nextlevel.remove(d);
                            node2goods.remove(d);
                            node2ks.remove(d);
                            intermediate.remove(d);
                            removedPending.add(d);
                            saveState(datatout, "Remove Pending:" + leveli, toGraph(new HashSet<String>(nextlevel), noisyGraph, pushed, analyzed, removedPending, removedNoisy, removedEdges, roots, leafNodes, intermediate));

                            for (String h : new HashSet<>(passing.keySet())) {
                                passing.remove(h, d);
                            }
                            for (String h : new HashSet<>(passingk.keySet())) {
                                passingk.remove(h, d);
                            }
                            saved.remove(d);
                            analyzed.remove(d);
                            if (debug) {

                                bw.write("$REMOVE-NODE PENDING LEAF: " + d + "\n");
                            }
                            if (additional) {
                                System.out.println("Rmv. leaf:" + d);
                            }
                            for (Integer i : new HashSet<Integer>(pushed.keySet())) {
                                pushed.remove(i, d);
                            }
                        }

                    } while (removed);

                    System.out.println("REMOVING PENDING....");
                    do {
                        removed = false;
                        Set<String> toremove = new HashSet<>();
                        for (String h : noisyGraph.vertexSet()) {
                            if (!roots.contains(h) && noisyGraph.incomingEdgesOf(h).isEmpty()) {
                                removed = true;
                                toremove.add(h);

                            }
                        }
                        for (String d : toremove) {
                            removedEdges.addAll(noisyGraph.incomingEdgesOf(d));
                            removedEdges.addAll(noisyGraph.outgoingEdgesOf(d));
                            removedPending.add(d);
                            saveState(datatout, "Remove Pending:" + leveli, toGraph(new HashSet<String>(nextlevel), noisyGraph, pushed, analyzed, removedPending, removedNoisy, removedEdges, roots, leafNodes, intermediate));
                            noisyGraph.removeVertex(d);
                            nextlevel.remove(d);
                            saved.remove(d);
                            analyzed.remove(d);
                            intermediate.remove(d);
                            leafNodes.remove(d);
                            if (debug) {

                                bw.write("$REMOVE-NODE PENDING ROOT: " + d + "\n");
                            }
                            if (additional) {
                                System.out.println("Rmv. root:" + d);
                            }
                            for (Integer i : new HashSet<Integer>(pushed.keySet())) {
                                pushed.remove(i, d);

                            }
                            node2goods.remove(d);
                            for (String h : new HashSet<>(passing.keySet())) {
                                passing.remove(h, d);
                            }

                            node2ks.remove(d);
                            for (String h : new HashSet<>(passingk.keySet())) {
                                passingk.remove(h, d);
                            }
                        }
                    } while (removed);

                    saved.addAll(nextlevel);
                }//ENF IF FIRST LEVEL

                // ora si effettua la rimozione di nodi non essenziali
                {
                    Set<String> kprivileged = new HashSet<>(intermediate);
                    kprivileged.retainAll(nextlevel);
                    Set<String> goodessential2 = new HashSet<>(essential);
                    goodessential2.addAll(leafNodes);

                    for (String g : leafNodes) {
                        if (passing.containsKey(g) && passing.get(g).size() == 1) {
                            essential.addAll(passing.get(g));
                        }
                    }

                    for (String k : intermediate) {
                        if (passingk.containsKey(k) && passingk.get(k).size() == 1) {
                            essential.addAll(passingk.get(k));

                        }
                    }

                    essential.add(rootLabel);
                    essential.addAll(intermediate);
                    essential.addAll(leafNodes);
                    if (debug) {
                        bw.write("$UNDELETABLE:" + essential + "\n");
                        bw.write("$TOGROUNDTABLE:" + passing + "\n");
                        bw.write("$TOINTERMEDIATETABLE:" + passingk + "\n");
                        System.out.println("ESSE:" + essential);
                        System.out.println("PASSSING:" + passing);
                        System.out.println("PASSSINGK:" + passingk);
                    }

                    ScreenUpdate.cleanScreen();
                    ScreenUpdate.printText("level " + leveli + " size:" + nextlevel.size(), 1, 1);
                    //saved.addAll(nextlevel);
                    DoubleCounter<String> srank = new DoubleCounter<>();
                    Set<String> res = new HashSet<>();

                    if (!removeOnlyCycles) {
                        for (String n : new HashSet<>(nextlevel)) {
                            Double r = getRank(n, leafNodes, intermediate, essential, node2goods, node2ks, analyzed);
                            if (!roots.contains(n) && r == Double.MAX_VALUE) {
                                if (additional) {
                                    Set<String> incomingofN = new HashSet<>();
                                    Set<String> outgoingofN = new HashSet<>();
                                    for (DefaultEdge de : noisyGraph.incomingEdgesOf(n)) {
                                        incomingofN.add(noisyGraph.getEdgeSource(de));
                                    }
                                    for (DefaultEdge de : noisyGraph.outgoingEdgesOf(n)) {
                                        outgoingofN.add(noisyGraph.getEdgeTarget(de));
                                    }
                                    noisyGraph.removeVertex(n);
                                    analyzed.remove(n);
                                    Set<String> tocheck = new HashSet<>();
                                    for (String hj : leafNodes) {
                                        if (passing.containsEntry(hj, n)) {
                                            tocheck.add(hj);
                                        }
                                    }
                                    if (!tocheck.isEmpty() && quickCheck(noisyGraph, rootLabel, tocheck)) {
                                        noisyGraph.addVertex(n);
                                        analyzed.add(n);
                                        for (String in : incomingofN) {
                                            noisyGraph.addEdge(in, n);
                                        }
                                        for (String on : outgoingofN) {
                                            noisyGraph.addEdge(n, on);
                                        }
                                        System.out.println("********* UNREMOVABLE NODE:" + n + " " + essential.contains(n) + " " + node2goods.get(n) + " " + node2ks.get(n) + " " + srank.get(n));
                                        essential.add(n);
                                        somethingchanged = true;
                                        continue;
                                    }
                                }
                                removedEdges.addAll(noisyGraph.incomingEdgesOf(n));
                                removedEdges.addAll(noisyGraph.outgoingEdgesOf(n));

                                noisyGraph.removeVertex(n);
                                analyzed.remove(n);
                                removedNoisy.add(n);
                                saveState(datatout, "Remove Unessential:" + leveli, toGraph(new HashSet<String>(nextlevel), noisyGraph, pushed, analyzed, removedPending, removedNoisy, removedEdges, roots, leafNodes, intermediate));
                                if (debug) {
                                    bw.write("$REMOVE-NODE NOT NECESSARY: " + n + "\n");
                                }
                                nextlevel.remove(n);
                                saved.remove(n);
                                res.add(n);
                                somethingchanged = true;
                                node2goods.remove(n);
                                node2ks.remove(n);

                                for (String h : new HashSet<>(passing.keySet())) {
                                    if (passing.containsEntry(h, n)) {
                                        passing.remove(h, n);
                                    }

                                }
                                for (String h : new HashSet<>(passingk.keySet())) {
                                    passingk.remove(h, n);
                                }

                                for (Integer li : new HashSet<>(pushed.keySet())) {
                                    pushed.remove(li, n);
                                }
                            } else {

                                srank.count(n, r);
                            }
                        }
                    }
                    if (debug) {
                        for (String j : srank.getSortedElements()) {
                            bw.write("$RANK: " + j + " " + srank.get(j) + "\n");
                        }
                        System.out.println("srank:" + srank);
                    }

                    if (!somethingchanged) {

                        List<String> tops = srank.getTops();
                        for (String s : tops) {
                            if (srank.get(s) > 0 && !roots.contains(s)) {
                                if (additional) {
                                    Set<String> incomingofN = new HashSet<>();
                                    Set<String> outgoingofN = new HashSet<>();
                                    for (DefaultEdge de : noisyGraph.incomingEdgesOf(s)) {
                                        incomingofN.add(noisyGraph.getEdgeSource(de));
                                    }
                                    for (DefaultEdge de : noisyGraph.outgoingEdgesOf(s)) {
                                        outgoingofN.add(noisyGraph.getEdgeTarget(de));
                                    }
                                    noisyGraph.removeVertex(s);
                                    analyzed.remove(s);
                                    Set<String> tocheck = new HashSet<>();
                                    for (String hj : leafNodes) {
                                        if (passing.containsEntry(hj, s)) {
                                            tocheck.add(hj);
                                        }
                                    }
                                    if (!tocheck.isEmpty() && quickCheck(noisyGraph, rootLabel, tocheck)) {
                                        noisyGraph.addVertex(s);
                                        analyzed.add(s);
                                        for (String in : incomingofN) {
                                            noisyGraph.addEdge(in, s);
                                        }
                                        for (String on : outgoingofN) {
                                            noisyGraph.addEdge(s, on);
                                        }
                                        System.out.println("********* UNREMOVABLE NODE:" + s + " " + essential.contains(s) + " " + node2goods.get(s) + " " + node2ks.get(s) + " " + srank.get(s));
                                        essential.add(s);
                                        somethingchanged = true;
                                        continue;
                                    }
                                }
                                removedNoisy.add(s);
                                removedEdges.addAll(noisyGraph.incomingEdgesOf(s));
                                removedEdges.addAll(noisyGraph.outgoingEdgesOf(s));
                                saveState(datatout, "Remove Unessential:" + leveli, toGraph(new HashSet<String>(nextlevel), noisyGraph, pushed, analyzed, removedPending, removedNoisy, removedEdges, roots, leafNodes, intermediate));
                                noisyGraph.removeVertex(s);
                                saved.remove(s);
                                analyzed.remove(s);
                                somethingchanged = true;
                                noisyGraph.removeVertex(s);
                                nextlevel.remove(s);
                                res.add(s);
                                if (debug) {
                                    bw.write("$REMOVE-NODE NOISY: " + s + "\n");
                                }
                                node2goods.remove(s);
                                node2ks.remove(s);
                                for (String h : new HashSet<>(passing.keySet())) {
                                    passing.remove(h, s);
                                }
                                for (String h : new HashSet<>(passingk.keySet())) {
                                    passingk.remove(h, s);
                                }
                                for (Integer li : new HashSet<>(pushed.keySet())) {
                                    pushed.remove(li, s);
                                }
                                somethingchanged = true;
                                break;
                            }
                        }
                    }

                    // ottimizza la memoria
                    for (Integer li : new HashSet<>(pushed.keySet())) {
                        for (String s : res) {
                            pushed.remove(li, s);
                        }
                    }
                }

            } while (somethingchanged);

            passing.clear();
            passingk.clear();
            level = nextlevel;

            max = 0;
            for (Integer m : pushed.keySet()) {
                if (m > max && pushed.size() > 0 && pushed.get(m) != null && !pushed.get(m).isEmpty() /*&& leveli < 4000*/) {
                    max = m;
                }
            }
            if (debug) {
                bw.write("$SAVED:" + saved.size() + "\n");
                bw.write("$MAX: " + max + "\n");
                bw.write("$CONTAINS ROOT:" + analyzed.contains(rootLabel) + "\n");
                bw.write("$|G.V|: " + noisyGraph.vertexSet().size() + "\n");
                bw.write("$|G.E|: " + noisyGraph.edgeSet().size() + "\n");
            }
            bw.flush();
            if (leveli % 20 == 0) {
                saveGraph(noisyGraph, "data/piero/" + rootLabel + "_tpiero_" + leveli + ".tsv");
            }
        } while (!analyzed.contains(rootLabel) || max > leveli);

        Set<String> appv = new HashSet<>(noisyGraph.vertexSet());
        for (String a : appv) {
            if (!saved.contains(a)) {
                noisyGraph.removeVertex(a);
                analyzed.remove(a);
            }
        }

        boolean removed;
        do {
            removed = false;
            Set<String> toremove = new HashSet<>();
            for (String h : noisyGraph.vertexSet()) {
                if (!leafNodes.contains(h) && noisyGraph.outgoingEdgesOf(h).isEmpty()) {
                    removed = true;
                    toremove.add(h);
                }
            }
            for (String d : toremove) {
                removedPending.add(d);
                removedEdges.addAll(noisyGraph.incomingEdgesOf(d));
                removedEdges.addAll(noisyGraph.outgoingEdgesOf(d));
                saveState(datatout, "Remove Pending:" + leveli, toGraph(new HashSet<String>(nextlevel), noisyGraph, pushed, analyzed, removedPending, removedNoisy, removedEdges, roots, leafNodes, intermediate));
                removedEdges.addAll(noisyGraph.incomingEdgesOf(d));
                removedEdges.addAll(noisyGraph.outgoingEdgesOf(d));
                noisyGraph.removeVertex(d);
                intermediate.remove(d);
                analyzed.remove(d);
                if (debug) {

                    bw.write("$REMOVE-NODE PENDING LEAF:" + d + "\n");
                }
                if (additional) {
                    System.out.println("Rmv. leaf:" + d);
                }
            }
            if (!sanityCheck(noisyGraph, rootLabel, leafNodes, intermediate)) {
                System.exit(0);
            }
        } while (removed);

        do {
            removed = false;
            Set<String> toremove = new HashSet<>();
            for (String h : noisyGraph.vertexSet()) {
                if (!intermediate.contains(h) && !roots.contains(h) && noisyGraph.incomingEdgesOf(h).isEmpty()) {
                    removed = true;
                    toremove.add(h);
                }
            }
            for (String d : toremove) {
                removedPending.add(d);
                removedEdges.addAll(noisyGraph.incomingEdgesOf(d));
                removedEdges.addAll(noisyGraph.outgoingEdgesOf(d));
                removedEdges.addAll(noisyGraph.incomingEdgesOf(d));
                removedEdges.addAll(noisyGraph.outgoingEdgesOf(d));
                saveState(datatout, "Remove Pending:" + leveli, toGraph(new HashSet<String>(nextlevel), noisyGraph, pushed, analyzed, removedPending, removedNoisy, removedEdges, roots, leafNodes, intermediate));
                noisyGraph.removeVertex(d);
                analyzed.remove(d);
                if (debug) {

                    bw.write("$REMOVE-NODE PENDING ROOT:" + d + "\n");
                }
                if (additional) {
                    System.out.println("Rmv. root:" + d);
                }
            }
        } while (removed);

        if (debug) {
            for (DefaultEdge de : noisyGraph.edgeSet()) {
                bw.write("$ACTUAL: " + de + "\n");
            }
            bw.write("END\n");
            for (DefaultEdge de : noisyGraph.edgeSet()) {
                bw.write("$ACTUAL: " + de + "\n");
            }
            bw.write("END\n");

        }

        bw.close();

        System.out.println("Max postponed:" + postponedcounts.getTopValue());
        analyzed.clear();
        saveState(datatout, "Final graph", toGraph(new HashSet<String>(), noisyGraph, pushed, analyzed, removedPending, removedNoisy, removedEdges, roots, leafNodes, intermediate));
        double total = 0;
        for (String s : postponedcounts.keySet()) {
            total += postponedcounts.get(s);
        }
        System.out.println("TOTAL:" + total);
        System.out.println("AVERAGE:" + (double) total / (double) (postponedcounts.keySet().size()));
    }

    public static boolean sanityCheck(DefaultDirectedGraph<String, DefaultEdge> result, String root, Set<String> good, Set<String> K) throws IOException {
        boolean wrong = false;
        if (!result.containsVertex(root)) {
            System.out.println(" Graph must contains:" + root);
            wrong = true;
        }

        for (String g : good) {
            if (!g.equals(root)) {
                if (!result.containsVertex(g)) {
                    System.out.println(" Graph must contains:" + g);
                    wrong = true;

                } else {
                    Double cost = Double.POSITIVE_INFINITY;
                    List<DefaultEdge> bf = DijkstraShortestPath.findPathBetween(result, root, g);
                    if (bf != null && !bf.isEmpty()) {
                        cost = new Double(bf.size());
                    }

                    if (cost == Double.POSITIVE_INFINITY) {
                        System.out.println(" WRONG G:" + g);
                        saveGraph(result, "data/Events/WRONG" + "_" + g + ".tsv");
                        for (DefaultEdge de : result.incomingEdgesOf(g)) {
                            System.out.println(" INCOMING OF G:" + de);
                        }
                        wrong = true;
                    }
                }
            }
        }

        return !wrong;
    }

    public static boolean quickCheck(DefaultDirectedGraph<String, DefaultEdge> result, String root, Set<String> good) {

        for (String g : good) {
            if (!g.equals(root)) {

                Double cost = Double.POSITIVE_INFINITY;
                List<DefaultEdge> bf = DijkstraShortestPath.findPathBetween(result, root, g);
                if (bf != null && !bf.isEmpty()) {
                    cost = new Double(bf.size());
                }
                if (cost == Double.POSITIVE_INFINITY) {

                    return true;
                }
            }
        }

        return false;
    }

    public static Double getRank(String n, Set<String> good, Set<String> K, Set<String> essential, IntegerCounter<String> node2goods, IntegerCounter<String> node2ks, Set<String> analyzed) throws IOException {
        double goodv = 0.0;
        double kv = 0.0;

        if (good != null && good.contains(n)) {
            return 0.0;
        }
        // se e' essenziale non lo elimina
        if (essential != null && essential.contains(n)) {
            return 0.0;
        }
        if (analyzed != null && analyzed.contains(n)) {
            return 0.0;
        }

        if (node2goods.keySet().contains(n)) {
            Integer i = node2goods.get(n);
            if (i != null) {
                goodv = i;
            }
        }

        if (goodv == 0) {
            return Double.MAX_VALUE;
        }

        Double arank = (1.0 - ((double) (goodv) / (double) (good.size())));
        arank = (double) arank + (double) 1.0;
        return arank;

    }

    private static DefaultEdge removeCycle(boolean additional, int leveli, String s, ArrayList<ArrayList<DefaultEdge>> p_r_s, String so,/* ArrayList<ArrayList<DefaultEdge>> p_r_so,*/ DefaultDirectedGraph<String, DefaultEdge> result, String root, String label, BufferedWriter bw, /*BellmanFordShortestPath bf_root,*/ Set<String> analyzed, Set<String> protec, boolean debug, Set<String> goods) throws IOException {

        String C = s;
        String B = so;
        DefaultEdge removed = null;

        //   int iter = 0;
        boolean ifound = true;

        List<DefaultEdge> bfcb = null;
        while (result.containsEdge(B, C) && result.containsVertex(C) && result.containsVertex(B) && !C.equals(B) && (bfcb = BellmanFordShortestPath.findPathBetween(result, C, B)) != null) {
            //  iter++;
            bfcb.add(result.getEdge(B, C));

            if (debug) {
                try {

                    bw.write("$" + label + "-PASSING-THROUGH :" + C + "->" + B + "\n");

                    bw.write("$" + label + "-CYCLE :" + bfcb + "\n");
                } catch (Exception e) {
                }
                System.out.println("E=" + B + "->" + C + "\n");
                System.out.println("CYCLE:" + bfcb);
            }
            ifound = false;

            for (int zu = bfcb.size() - 1; zu >= 0; zu--) //  for (int zu =0; zu <bfcb.size() ; zu++) 
            {
                DefaultEdge dez = bfcb.get(zu);
                String k = result.getEdgeSource(dez);
                String kt = result.getEdgeTarget(dez);
                if (p_r_s == null) {
                    p_r_s = getAllPathFromK(s, result, protec);
                }
                if (debug) {
                    try {
                        bw.write("$" + label + "-CANDIDATE :" + dez + "\n");
                        bw.write("$" + label + "-CANDIDATE-ROOT-SOURCE :" + p_r_s + "\n");
                    } catch (Exception e) {
                    }
                }
                if (!edgeContained(dez, p_r_s)) {

                    result.removeAllEdges(k, kt);
                    removed = dez;
                    if (additional && !sanityCheck(result, root, goods, protec)) {
                        result.addEdge(k, kt);
                        System.out.println("**** EDGE ESSENTIAL CAN NOT BE REMOVED!!!! *****" + k + " " + kt);
                        removed = null;
                        continue;
                    }

                    if (result.getEdge(B, C) == result.getEdge(k, kt)) {
                        System.out.println("************** REMOVING PASSING:" + B + "->" + C);
                    }
                    if (zu == 0) {
                        System.out.println("************** REMOVING c_1->c_2:" + k + "->" + kt);

                    }
                    ifound = true;
                    if (debug) {
                        try {
                            bw.write("$REMOVE-EDGE " + label + " " + zu + " : " + k + "->" + kt + "\n");
                            System.out.println("REMOVED EDGE:" + k + "->" + kt);
                        } catch (Exception e) {
                        }
                    }
                    break;
                }

            }

            if (!ifound) {
                System.out.println("ERR: " + label + ":" + B + "->" + C);
                System.exit(0);
            }

        }//while
        return removed;
    }

    public static boolean edgeContained(DefaultEdge de, ArrayList<ArrayList<DefaultEdge>> paths) {
        if (paths == null) {
            return false;
        }
        for (ArrayList<DefaultEdge> path : paths) {
            if (!path.isEmpty() && path.contains(de)) {
                return true;
            }
        }
        return false;
    }

    public static ArrayList<ArrayList<DefaultEdge>> getAllPathFromK(String target, DefaultDirectedGraph<String, DefaultEdge> result, Set<String> K) {

        ArrayList<ArrayList<DefaultEdge>> res = new ArrayList<>();
        for (String k : K) {
            if (!k.equals(target)) {
                ArrayList<DefaultEdge> p_k = (ArrayList<DefaultEdge>) BellmanFordShortestPath.findPathBetween(result, k, target);
                if (p_k != null) {
                    res.add(p_k);
                }
            }
        }

        return res;
    }

    public static DefaultEdge removeCyclesOverEdge(boolean additional, int leveli, String label, String s, String t, DefaultDirectedGraph<String, DefaultEdge> result, String root, /*BellmanFordShortestPath bfr,*/ BufferedWriter bw, Set<String> protec, Set<String> analyzed, boolean debug, Set<String> good) throws IOException {
        ArrayList<ArrayList<DefaultEdge>> p_r_s = null;
        return removeCycle(additional, leveli, s, p_r_s, t, result, root, label, bw,/* bfr,*/ analyzed, protec, debug, good);

    }

    public static void removeIsolated(DefaultDirectedGraph<String, DefaultEdge> graph) {

        for (String s : new HashSet<>(graph.vertexSet())) {
            if (graph.incomingEdgesOf(s).isEmpty() && graph.outgoingEdgesOf(s).isEmpty()) {
                graph.removeVertex(s);
            }
        }

        for (String s : new HashSet<>(graph.vertexSet())) {
            graph.removeAllEdges(s, s);
        }
    }

    public static void removeUnConnected(DefaultDirectedGraph<String, DefaultEdge> original, String root, Set<String> good, Set<String> K) {
        BellmanFordShortestPath<String, DefaultEdge> bf = new BellmanFordShortestPath(original, root);
        for (String v : new HashSet<>(K)) {
            if (v.equals(root)) {
                continue;
            }
            if (!original.containsVertex(v)) {
                System.out.println("Removing unconnected" + v + "from K");
                K.remove(v);
                continue;

            }
            Double cost = bf.getCost(v);
            if (cost.equals(Double.POSITIVE_INFINITY)) {
                System.out.println("Removing unconnected" + v + "from GOODS");
                K.remove(v);
            }

        }
        for (String v : new HashSet<>(good)) {
            if (v.equals(root)) {
                continue;
            }
            if (!original.containsVertex(v)) {
                good.remove(v);
                System.out.println("Removing unconnected" + v + "from GOODS");
                continue;
            }
            Double cost = bf.getCost(v);
            if (cost.equals(Double.POSITIVE_INFINITY)) {
                System.out.println("Removing unconnected" + v + "from GOODS");
                good.remove(v);
            }

        }
    }

    public static void saveGraph(DefaultDirectedGraph<String, DefaultEdge> graph, String filename) throws IOException {
        int id = 0;
        try (BufferedWriter bw = Files.getBufferedWriter(filename)) {
            for (DefaultEdge de : graph.edgeSet()) {
                id++;
                bw.write(id + "\t" + graph.getEdgeTarget(de) + "\t" + graph.getEdgeSource(de) + "\n");
                
            }
        }

    }

    public static void saveState(String filename, String description, String graph) throws IOException {
        BufferedWriter bw = Files.getBufferedWriter(filename + "_" + dataoutcounter + ".dat");
        bw.write(description + "\n");
        bw.write(graph);
        bw.close();
        dataoutcounter++;

    }

    public static String toGraph(Set<String> layer, DefaultDirectedGraph<String, DefaultEdge> graph, Multimap<Integer, String> pushed, Set<String> analyzed, Set<String> removedPending, Set<String> removedNoisy, Set<DefaultEdge> removededges, Set<String> roots, Set<String> leafs, Set<String> intermediate) {
        StringBuilder sb = new StringBuilder();
        HashMap<String, String> n2father = new HashMap();
        Set<String> additionalnodes = new HashSet<>();
        additionalnodes.add("Traversed");
        additionalnodes.add("Removed Pending");
        additionalnodes.add("Removed Noisy");
        additionalnodes.add("Current Layer");
        additionalnodes.add("Root");
        for (String s : roots) {
            n2father.put(s, "Root");
            if (layer.contains(s)) {
                n2father.put("Root", "Current Layer");
            }
        }
        for (int s : pushed.keySet()) {

            sb.append("v\t").append("V_" + s).append("\n");
            for (String r : pushed.get(s)) {
                if (roots.contains(r)) {
                    n2father.put("Root", "V_" + s);
                } else {
                    n2father.put(r, "V_" + s);
                }

            }
        }
        for (String s : layer) {

            if (roots.contains(s)) {
                n2father.put("Root", "Current Layer");
                n2father.put(s, "Root");
            } else {
                n2father.put(s, "Current Layer");
            }
        }

        for (String s : analyzed) {
            n2father.put(s, "Traversed");
            if (roots.contains(s)) {
                n2father.put("Root", "Traversed");
                n2father.put(s, "Root");
            }
        }
        for (String s : removedPending) {
            n2father.put(s, "Removed Pending");
            additionalnodes.add(s);
        }
        for (String s : removedNoisy) {
            n2father.put(s, "Removed Noisy");
            additionalnodes.add(s);
        }
        for (String s : additionalnodes) {
            String father = "undefined";
            if (n2father.containsKey(s)) {
                father = n2father.get(s);
            }
           // conta quante un volte un nodo e' father

            if (n2father.containsValue(s) || n2father.containsKey(s)) {
                sb.append("v\t").append(s).append("\t").append(father).append("\t").append("0").append("\n");
            }
        }
        for (String s : graph.vertexSet()) {
            String p = "0";
            if (leafs.contains(s)) {
                p = "l";
            }
            if (intermediate.contains(s)) {
                p = "i";
            }
            if (roots.contains(s)) {
                p = "r";
            }
            String father = "undefined";
            if (n2father.containsKey(s)) {
                father = n2father.get(s);
            }
            sb.append("v\t").append(s).append("\t").append(father).append("\t").append(p).append("\n");
        }
        for (DefaultEdge de : graph.edgeSet()) {
            if (graph.getEdgeTarget(de) == null
                    || graph.getEdgeSource(de) == null
                    || removedPending.contains(graph.getEdgeTarget(de))
                    || removedNoisy.contains(graph.getEdgeTarget(de))
                    || removedPending.contains(graph.getEdgeSource(de))
                    || removedNoisy.contains(graph.getEdgeSource(de))
                    || (n2father.containsKey(graph.getEdgeSource(de)) && n2father.get(graph.getEdgeSource(de)).startsWith("Remo"))) {
                removededges.add(de);
            } else {
                sb.append("e\t").append(graph.getEdgeSource(de)).append("\t").append(graph.getEdgeTarget(de)).append("\n");
            }
        }
        for (DefaultEdge de : removededges) {
            sb.append("re\t").append(graph.getEdgeSource(de)).append("\t").append(graph.getEdgeTarget(de)).append("\n");
        }

        return sb.toString();
    }
}
